/*====================================================================
:::
::: Epson.c
:::
::: Generic TealPrint Text Print Driver
::: (c)1999-2000 TealPoint Software All rights reserved
:::
::: This source code may only be used to develop a printer driver
::: for TealPrint.  This code may only be used to produce a freely-
::: distributable program.
:::
====================================================================*/

#pragma pack(2)

#include <Common.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>

#include "..\\tealprnt.h"
#include "utils.c"

#define MIN(a,b) ((a)<(b)?(a):(b))

//
// Constants
//

#define CREATOR_ID			'TPEp'		// This should be unique per driver

#define ID_FTR_irda_state	33333
#define PREFS_VERSION		14
#define NUM_PAPER_SIZES		8

//
// Resources
//

#define ID_FORM_main 						1000
#define ID_BUTN_MAIN_test					1001
#define ID_BUTN_MAIN_settings				1005
#define ID_FORM_settings					1100
#define ID_FIELD_margin_left				1156
#define ID_FIELD_margin_top				1158
													
#define ID_TRIG_settings_baud				1140
#define ID_LIST_settings_baud				1141
#define ID_CHECK_settings_rts_cts		1142
#define ID_CHECK_settings_xon_xoff		1143
#define ID_PUSH_settings_parityoff		1144
#define ID_PUSH_settings_parityodd		1145
#define ID_PUSH_settings_parityeven		1146
#define ID_PUSH_settings_linecrlf		1152
#define ID_PUSH_settings_linecr			1153
#define ID_PUSH_settings_linelf			1154
#define ID_CHECK_settings_sendslow		1147
#define ID_CHECK_settings_status			1161
#define ID_CHECK_settings_beepstatus	1162
#define ID_TRIG_settings_charset			1171
#define ID_LIST_settings_charset			1172

#define ID_FORM_info							1200

#define ID_ALERT_out_of_memory			1000
#define ID_ALERT_error_in_use				1001
#define ID_ALERT_error_find				1002
#define ID_ALERT_error_settings			1003
#define ID_ALERT_error_timed_out			1004
#define ID_ALERT_error_checking_port 	1005
#define ID_ALERT_error_line_error		1006
#define ID_ALERT_error_read_error		1007

#define	ID_STRING_driver_name			1000
#define	ID_STRING_text_mode0				1001
#define	ID_STRING_graphic_mode0			1030
#define	ID_STRING_text_remap0			2000

#define	ID_STRING_paper_size0 			1002
#define	ID_STRING_paper_size1 			1003
#define	ID_STRING_paper_size2 			1004
#define	ID_STRING_paper_size3 			1005
#define	ID_STRING_paper_size4 			1006
#define	ID_STRING_paper_size5 			1007
#define	ID_STRING_paper_size6 			1008
#define	ID_STRING_paper_size7 			1009
#define	ID_STRING_paper_size8 			1010
#define	ID_STRING_error		 			1011
#define	ID_STRING_print_start 			1012
#define	ID_STRING_test0		 			1013
#define	ID_STRING_test1		 			1014
#define	ID_STRING_test2		 			1015
#define	ID_STRING_test3		 			1016
#define	ID_STRING_ir_discover 			1017
#define	ID_STRING_ir_irlap	 			1018
#define	ID_STRING_ir_irlmp	 			1019
#define	ID_STRING_ir_datareq	 			1020
#define	ID_STRING_ir_timeout	 			1021
#define	ID_STRING_ir_noconn	 			1022
#define	ID_STRING_ir_disconnect			1023
#define	ID_STRING_ir_busy					1024
#define	ID_STRING_ir_noreply				1025
#define	ID_STRING_ir_noport				1026
#define	ID_STRING_printk		 			1031

#define MIN_MARGIN_b	250L
#define MIN_MARGIN_r	250L
#define X_CPI	10
#define Y_CPI	 6
#define X_DPI	60
#define Y_DPI	72
#define MAX_PAPER_width	8500L

//
// Globals
//

	typedef struct {

		int margin_t[NUM_PAPER_SIZES];
		int margin_b[NUM_PAPER_SIZES];
		int margin_l[NUM_PAPER_SIZES];
		int margin_r[NUM_PAPER_SIZES];

		int paper_w[NUM_PAPER_SIZES];
		int paper_h[NUM_PAPER_SIZES];

		long baud_rate;
		int  rts_cts;
		int  xon_xoff;
		int  parity;
		int  linebreak;
		int  sendslow;
		int	margin_left;
		int	margin_top;
		int	infared;

		int	show_status;
		int	beep_status;
		int	charset;

	} SavePrefsType;

//
// Prototypes
//

	#define NULL	0

	void  app_Run			( Ptr command_data, Word launch_flags );
	void	app_Terminate	( void );
	int	app_Initialize	( void );
	void	app_SetSettings( int internal );

	void  debug_Line		( int col, int row );

	long  PrintTest		( long mode );
	long  PrintTextData  ( int cols, int lines, int blank_cols, int blank_lines, void get_line( char **data, int line, int *offset, int *char_count, void *app_data ), void *app_data, long print_flags );
	long	PrintBitmapData( int width, int height, int blank_cols, int blank_lines, void (*get_line)( unsigned char **data, int line, int pixel_offset, int pixel_count, void *app_data, unsigned char bgcolor ), void *app_data, long flags );
	long  PrintEjectPaper( void );

	void *GetObjectPtr( FormPtr form, int index );


/*======================================================================
:::
::: PilotMain()
:::
::: The Main entry point
:::
======================================================================*/

DWord PilotMain(Word cmd, Ptr command_data, Word launch_flags)
{
	switch (cmd) {
		//==========================================
		//
		// Run Normally
		//
		// This is the only routine in which you
		// can use global variables
		//
		//==========================================

		case sysAppLaunchCmdNormalLaunch:
			app_Run( command_data, launch_flags );
			break;

		//==========================================
		//
		// Get Driver Info
		//
		// Return basic information about the 
		// driver, including supported modes
		//
		//==========================================

		case DRIVER_CMD_GetDriverInfo:
			{
				DriverInfoType *pdi = (DriverInfoType *)command_data; 

				pdi->version = 1;
				pdi->revision = 0;
				pdi->name = util_GetResourceString( ID_STRING_driver_name );
				pdi->flags = DRIVER_FLAG_supports_text     | 
								 DRIVER_FLAG_supports_graphics | 
								 DRIVER_FLAG_has_settings      | 
								 DRIVER_FLAG_has_info;
				pdi->num_gfxmodes = 1;
				pdi->num_txtmodes = 1;
				pdi->num_paper_sizes = NUM_PAPER_SIZES;
				pdi->retval = 0;
			}
			return 0;

		//==========================================
		//
		// Get Graphic Mode Info
		//
		// Get information about the passed 
		// graphics mode including resolution
		//
		//==========================================

		case DRIVER_CMD_GetGModeInfo:

			{
				DriverGModeInfoType *pmi = (DriverGModeInfoType *)command_data; 

				pmi->name = util_GetResourceString( ID_STRING_graphic_mode0 );
				pmi->flags = 0;
				pmi->supported_modes = DRIVER_GFXMODE_1bit;

				pmi->y_dpi = Y_DPI;					// Return resolution.  Fixed in this example
				pmi->x_dpi = X_DPI;
				pmi->x_pixels = 9*pmi->x_dpi;		// Maximum paper width in pixels.. faked for now
				pmi->y_pixels = 0;

				pmi->retval = 0;
			}
			return 0;

		//==========================================
		//
		// Get Text Mode Info
		//
		// Get information about the passed 
		// text mode including text density
		//
		//==========================================

		case DRIVER_CMD_GetTModeInfo:
			{
				DriverTModeInfoType *pmi = (DriverTModeInfoType *)command_data; 

				pmi->name = util_GetResourceString( ID_STRING_text_mode0 );
				pmi->flags = 0;
				pmi->supported_modes = DRIVER_TXTMODE_ascii;

				pmi->x_cpi = 10;		// return density (fixed for this driver)
				pmi->y_cpi = 6;
				pmi->retval = 0;
			}
			return 0;

		//==========================================
		//
		// Get Paper Info
		//
		// Return information about the passed 
		// paper size
		//
		//==========================================

		case DRIVER_CMD_GetPaperInfo:
			{
				DriverPaperInfoType *dpi = (DriverPaperInfoType *)command_data; 
				SavePrefsType prefs;

				prefs_Load( &prefs );

				dpi->flags = 0;
				dpi->retval = 0;

				// minimum margins

				dpi->t_mils_min = prefs.margin_top;
				dpi->b_mils_min = MIN_MARGIN_b;
				dpi->l_mils_min = prefs.margin_left;
				dpi->r_mils_min = MIN_MARGIN_r;

				// default margins

				dpi->t_mils_def = (dpi->paper_number >=2 && dpi->paper_number<=6 ) ? 250:1000;
				dpi->b_mils_def = (dpi->paper_number >=2 && dpi->paper_number<=6 ) ? 250:1000;
				dpi->l_mils_def = (dpi->paper_number >=2 && dpi->paper_number<=6 ) ? 250:1000;
				dpi->r_mils_def = (dpi->paper_number >=2 && dpi->paper_number<=6 ) ? 250:1000;

				// current margins

				dpi->t_mils_cur = prefs.margin_t[dpi->paper_number];
				dpi->b_mils_cur = prefs.margin_b[dpi->paper_number];
				dpi->l_mils_cur = prefs.margin_l[dpi->paper_number];
				dpi->r_mils_cur = prefs.margin_r[dpi->paper_number];

				// Paper sizes

				dpi->w_mils_tot = prefs.paper_w[dpi->paper_number];
				dpi->h_mils_tot = prefs.paper_h[dpi->paper_number];

				switch( dpi->paper_number ) {

					case 0:
						dpi->name = util_GetResourceString( ID_STRING_paper_size0 );
						break;

					case 1:
						dpi->name = util_GetResourceString( ID_STRING_paper_size1 );
						break;

					case 2:
						dpi->name = util_GetResourceString( ID_STRING_paper_size2 );
						break;

					case 3:
						dpi->name = util_GetResourceString( ID_STRING_paper_size3 );
						break;

					case 4:
						dpi->name = util_GetResourceString( ID_STRING_paper_size4 );
						break;

					case 5:
						dpi->name = util_GetResourceString( ID_STRING_paper_size5 );
						break;

					case 6:
						dpi->name = util_GetResourceString( ID_STRING_paper_size6 );
						break;

					case 7:
						dpi->name = util_GetResourceString( ID_STRING_paper_size7 );
						break;

					case 8:
						dpi->name = util_GetResourceString( ID_STRING_paper_size8 );
						break;

					default:
						dpi->name = util_GetResourceString( ID_STRING_error );
						dpi->retval = 0;
						break;
				}
			}
			return 0;

		//==========================================
		//
		// Set Paper Info
		//
		//==========================================

		case DRIVER_CMD_SetPaperInfo:
			{
				DriverPaperInfoType *dpi = (DriverPaperInfoType *)command_data; 
				SavePrefsType prefs;

				// current margins

				if (dpi->paper_number >=0 && dpi->paper_number < NUM_PAPER_SIZES) {
					prefs_Load( &prefs );

					prefs.margin_t[dpi->paper_number] = dpi->t_mils_cur;
					prefs.margin_b[dpi->paper_number] = dpi->b_mils_cur;
					prefs.margin_l[dpi->paper_number] = dpi->l_mils_cur;
					prefs.margin_r[dpi->paper_number] = dpi->r_mils_cur;

					// Paper sizes

					prefs.paper_w[dpi->paper_number] = dpi->w_mils_tot;
					prefs.paper_h[dpi->paper_number] = dpi->h_mils_tot;

					prefs_Save( &prefs );
					dpi->retval = 0;
				} else dpi->retval = -1;
			}
			return 0;

		//==========================================
		//
		// Show Info
		//
		// Bring up a dialog with information about
		// the current driver
		//
		//==========================================

		case DRIVER_CMD_ShowInfo:
			{
				FormPtr form;
				int retval;
				DmOpenRef openref;

				if (openref = DmOpenDatabaseByTypeCreator( 'appl', CREATOR_ID, dmModeReadOnly )) {

					form = FrmInitForm( ID_FORM_info );

					retval = FrmDoDialog( form );
					FrmDeleteForm( form );

					DmCloseDatabase( openref );
				}
			}
			return 0;

		//==========================================
		//
		// Set Settings
		//
		// Adjust settings for the current driver
		//
		//==========================================

		case DRIVER_CMD_SetSettings:
			app_SetSettings(false);
			return 0;

		//==========================================
		//
		// Print a Bitmap
		//
		//==========================================

		case DRIVER_CMD_PrintBitmap:
			{
				DriverPrintBitmapType *dpb = (DriverPrintBitmapType *)command_data; 
				dpb->retval = PrintBitmapData( dpb->width, dpb->height, dpb->blank_cols, dpb->blank_lines, dpb->getraster_func, dpb->app_data, dpb->print_flags );
			}
			return 0;

		//==========================================
		//
		// Print some raw text
		//
		//==========================================

		case DRIVER_CMD_PrintText:
			{
				DriverPrintTextType *dpt = (DriverPrintTextType *)command_data; 
	
				dpt->retval = PrintTextData( dpt->cols, dpt->lines, dpt->blank_cols, dpt->blank_lines, dpt->getline_func, dpt->app_data, dpt->print_flags );
				return 0;
			}

		//==========================================
		//
		// Eject Paper
		//
		//==========================================

		case DRIVER_CMD_EjectPaper:
			PrintEjectPaper();
			return 0;

		//==========================================
		//
		// Reset
		//
		//==========================================

		case DRIVER_CMD_HardReset:
		case DRIVER_CMD_SoftReset:
			return 0;
	}

	return 0;
}

/*------------------------------------------------
:::
::: GetObjectPtr()
:::
------------------------------------------------*/
void *GetObjectPtr( FormPtr form, int index )
{
	return FrmGetObjectPtr( form, FrmGetObjectIndex( form, index ) );
}


/*======================================================================
:::
::: app_Run()
:::
::: The Main Loop
:::
======================================================================*/

	int NullHandler() { return 0; }

void app_Run( Ptr command_data, Word launch_flags )
{
	short err;
	EventType evt;
	int pendown,pendir=0,delay=1,newdelay=1;
	FormPtr form;

	if (!app_Initialize()) {

		FrmGotoForm(ID_FORM_main);

		while(1) {

			EvtGetEvent(&evt, evtWaitForever);

			if (SysHandleEvent(&evt)) continue;
			if (MenuHandleEvent((void *)0, &evt, &err)) continue;

			//==================================
			//
			// Handle post-system events
			//
			//==================================

			switch (evt.eType) {

				//-------------------------------------------------
				//
				// Nil Event
				//
				//-------------------------------------------------

		      case nilEvent:
					break;

				//-------------------------------------------------
				//
				// Menu Events
				//
				//-------------------------------------------------

		      case menuEvent:

					switch( evt.data.menu.itemID ) {
					}
					break;

				//-------------------------------------------------
				//
				// Form Load Events
				//
				//-------------------------------------------------

				case frmLoadEvent:
					FrmSetActiveForm(form = FrmInitForm(evt.data.frmLoad.formID));
			      FrmSetEventHandler(form, (void *)NullHandler);
					break;

				//-------------------------------------------------
				//
				// Form Open Events
				//
				//-------------------------------------------------

				case frmOpenEvent:
					FrmDrawForm(FrmGetActiveForm());
					break;

				//-------------------------------------------------
				//
				// Control Select Events
				//
				//-------------------------------------------------

				case ctlSelectEvent:
					switch( evt.data.ctlSelect.controlID ) {
						case ID_BUTN_MAIN_test:
							PrintTest( 0 );
							break;
						case ID_BUTN_MAIN_settings:
							app_SetSettings(true);
							break;
					}
					break;

				//-------------------------------------------------
				//
				// Stop Event
				//
				//-------------------------------------------------

				case appStopEvent:
					app_Terminate();
					return;

				default:
			   	FrmDispatchEvent(&evt);
					break;
			}
		}

		app_Terminate();
	}
}

/*------------------------------------------------
:::
::: BusyLoop()
:::
::: Dummy input loop for Infared Code
:::
------------------------------------------------*/

int BusyLoop( void )
{
	EventType event;
	UInt error;

	EvtGetEvent(&event, 100);

	if (event.eType == appStopEvent) return -1;

   if (SysHandleEvent(&event)) return 0;
	if (MenuHandleEvent(0, &event, &error)) return 0;
   FrmHandleEvent(FrmGetActiveForm(),&event) ;
	return 0;
}


/*======================================================================
:::
::: app_Initialize()
:::
======================================================================*/

int app_Initialize( void )
{
	return 0;
}

/*======================================================================
:::
::: app_Terminate()
:::
======================================================================*/

void app_Terminate( void )
{
}

/*======================================================================
:::
::: app_SetSettings()
:::
======================================================================*/

void app_SetSettings( int internal )
{
	FormPtr form;
	int retval;
	DmOpenRef openref;
	SavePrefsType prefs;
	int index;

	if (internal || (openref = DmOpenDatabaseByTypeCreator( 'appl', CREATOR_ID, dmModeReadOnly ))) {

		prefs_Load( &prefs );
		if (prefs.infared) {
			index = 
				(prefs.baud_rate ==   9600) ? 6:
				(prefs.baud_rate == 115200) ? 8:7;
		} else {
			index = 
				(prefs.baud_rate ==      0) ? 0:
				(prefs.baud_rate ==   9600) ? 1:
				(prefs.baud_rate ==  19200) ? 2:
				(prefs.baud_rate ==  38400) ? 3:
				(prefs.baud_rate ==  57600) ? 4:5;
		}

		form = FrmInitForm( ID_FORM_settings );

 		LstSetSelection( GetObjectPtr( form, ID_LIST_settings_baud   ), index );
 		LstSetSelection( GetObjectPtr( form, ID_LIST_settings_charset), prefs.charset );

		CtlSetLabel( GetObjectPtr( form, ID_TRIG_settings_baud ),
			LstGetSelectionText( GetObjectPtr( form, ID_LIST_settings_baud ), index ) );

		CtlSetLabel( GetObjectPtr( form, ID_TRIG_settings_charset ),
			LstGetSelectionText( GetObjectPtr( form, ID_LIST_settings_charset ), prefs.charset ) );

 		CtlSetValue( GetObjectPtr( form, ID_CHECK_settings_sendslow), prefs.sendslow );
 		CtlSetValue( GetObjectPtr( form, ID_CHECK_settings_rts_cts ), prefs.rts_cts  );
 		CtlSetValue( GetObjectPtr( form, ID_CHECK_settings_xon_xoff), prefs.xon_xoff );
 		CtlSetValue( GetObjectPtr( form, ID_CHECK_settings_status  ), prefs.show_status );
 		CtlSetValue( GetObjectPtr( form, ID_CHECK_settings_beepstatus), prefs.beep_status );

 		CtlSetValue( GetObjectPtr( form, ID_PUSH_settings_parityoff  ), prefs.parity==0 );
 		CtlSetValue( GetObjectPtr( form, ID_PUSH_settings_parityodd  ), prefs.parity==1 );
 		CtlSetValue( GetObjectPtr( form, ID_PUSH_settings_parityeven ), prefs.parity==2 );

 		CtlSetValue( GetObjectPtr( form, ID_PUSH_settings_linecrlf  ), prefs.linebreak==0 );
 		CtlSetValue( GetObjectPtr( form, ID_PUSH_settings_linecr    ), prefs.linebreak==1 );
 		CtlSetValue( GetObjectPtr( form, ID_PUSH_settings_linelf    ), prefs.linebreak==2 );

		util_SetFieldNumber( form, ID_FIELD_margin_left, prefs.margin_left, 3, false );
		util_SetFieldNumber( form, ID_FIELD_margin_top,  prefs.margin_top,  3, false );

		retval = FrmDoDialog( form );

		prefs.margin_left = util_GetFieldNumber( form, ID_FIELD_margin_left, 3 );
		prefs.margin_top  = util_GetFieldNumber( form, ID_FIELD_margin_top,  3 );

		prefs.charset   = LstGetSelection( GetObjectPtr( form, ID_LIST_settings_charset ));
		prefs.baud_rate = LstGetSelection( GetObjectPtr( form, ID_LIST_settings_baud ));
		prefs.infared   = prefs.baud_rate > 5;
		prefs.baud_rate = 
			(prefs.baud_rate==0                      ) ?      0:
			(prefs.baud_rate==1 || prefs.baud_rate==6) ?   9600:
			(prefs.baud_rate==2                      ) ?  19200:
			(prefs.baud_rate==3                      ) ?  38400:
			(prefs.baud_rate==7 || prefs.baud_rate==4) ?  57600:115200;

 		prefs.rts_cts     = CtlGetValue( GetObjectPtr( form, ID_CHECK_settings_rts_cts ) );
 		prefs.xon_xoff    = CtlGetValue( GetObjectPtr( form, ID_CHECK_settings_xon_xoff) );
 		prefs.show_status = CtlGetValue( GetObjectPtr( form, ID_CHECK_settings_status  ) );
 		prefs.sendslow    = CtlGetValue( GetObjectPtr( form, ID_CHECK_settings_sendslow) );
 		prefs.beep_status = CtlGetValue( GetObjectPtr( form, ID_CHECK_settings_beepstatus  ) );
 		prefs.parity      = 
			CtlGetValue( GetObjectPtr( form, ID_PUSH_settings_parityeven  ) ) ? 2 :
			CtlGetValue( GetObjectPtr( form, ID_PUSH_settings_parityodd   ) ) ? 1:0;
 		prefs.linebreak = 
			CtlGetValue( GetObjectPtr( form, ID_PUSH_settings_linelf  ) ) ? 2 :
			CtlGetValue( GetObjectPtr( form, ID_PUSH_settings_linecr  ) ) ? 1:0;

		FrmDeleteForm( form );


		prefs_Save( &prefs );

		if (!internal) DmCloseDatabase( openref );
	}
}

/*======================================================================
:::
::: prefs_Load()
:::
======================================================================*/

int prefs_Load( SavePrefsType *prefs )
{
	int i;

	MemSet( prefs, sizeof(*prefs), 0 );

	prefs->margin_left = 250;
	prefs->margin_top  = 500;

	prefs->beep_status = true;
	prefs->show_status = true;

	prefs->baud_rate = 57600;
	prefs->xon_xoff = 1;
	prefs->rts_cts = 0;
	prefs->sendslow= 1;
	prefs->parity  = 0;
	prefs->linebreak = 0;

	for (i=0; i<NUM_PAPER_SIZES; i++) {
		prefs->margin_l[i] = 1000;
		prefs->margin_r[i] = 1000;
		prefs->margin_t[i] = 1000;
		prefs->margin_b[i] = 1000;
		prefs->paper_w[i] =  8500;
		prefs->paper_h[i] =  11000;
	}

	for (i=2; i<=6; i++) {
		prefs->margin_l[i] = 250;
		prefs->margin_r[i] = 250;
		prefs->margin_t[i] = 250;
		prefs->margin_b[i] = 250;
	}

	prefs->paper_w[0] =  8500;
	prefs->paper_h[0] = 11000;

	prefs->paper_w[1] =  8500;
	prefs->paper_h[1] = 14000;

	prefs->paper_w[2] = 9250;
	prefs->paper_h[2] = 4000;

	prefs->paper_w[3] = 4000;
	prefs->paper_h[3] = 9250;

	prefs->paper_w[4] = 4000;
	prefs->paper_h[4] = 6000;

	prefs->paper_w[5] = 5000;
	prefs->paper_h[5] = 7000;

	prefs->paper_w[6] = 8000;
	prefs->paper_h[6] = 10000;

	prefs->paper_w[7] = 8500;
	prefs->paper_h[7] = 17000;

	PrefGetAppPreferencesV10( CREATOR_ID, PREFS_VERSION, prefs, sizeof(*prefs) );
	return 0;
}

/*======================================================================
:::
::: prefs_Save()
:::
======================================================================*/

int prefs_Save( SavePrefsType *prefs )
{
	PrefSetAppPreferencesV10( CREATOR_ID, PREFS_VERSION, prefs, sizeof(*prefs) );

	return 0;
}

#define INFO_x	0
#define INFO_y	0
#define INFO_w	160
#define INFO_h	20

/*======================================================================
:::
::: status_Show()
:::
======================================================================*/
void status_Show( WindowType *window, char *string )
{
	if (window) {
		RectangleType rect;
		WindowType *savewin;
		int savefont;
	
		savefont = FntGetFont();
		FntSetFont( 0 );
	
		savewin = WinGetDrawWindow();
		WinSetDrawWindow( WinGetDisplayWindow() );
	
		rect.topLeft.x = INFO_x;	
		rect.topLeft.y = INFO_y;	
		rect.extent.x  = INFO_w;	
		rect.extent.y  = INFO_h;	
	
		WinEraseRectangle( &rect, 0 );
	
		rect.topLeft.x = INFO_x+2;	
		rect.topLeft.y = INFO_y+2;	
		rect.extent.x  = INFO_w-4;	
		rect.extent.y  = INFO_h-4;	
	
		WinDrawRectangleFrame( popupFrame, &rect );
	
		// Draw Text
	
		WinDrawChars( string, StrLen(string), INFO_x+10, INFO_y+4 );
	
		WinSetDrawWindow( savewin );
		FntSetFont(savefont);
	}
}

/*======================================================================
:::
::: status_Initialize()
:::
======================================================================*/
void status_Initialize( WindowType **winptr, SavePrefsType *prefs )
{
	UInt error;

	if (!prefs->show_status) {
		*winptr = 0;
		return;
	}

	if ( *winptr = WinCreateOffscreenWindow( INFO_w,INFO_h,screenFormat,&error )) {
		RectangleType rect;
		WindowType *savewin;

		savewin = WinGetDrawWindow();
		WinSetDrawWindow( WinGetDisplayWindow() );

		rect.topLeft.x = INFO_x;	
		rect.topLeft.y = INFO_y;	
		rect.extent.x  = INFO_w;	
		rect.extent.y  = INFO_h;	

		WinCopyRectangle( WinGetDisplayWindow(), *winptr, &rect, 0,0, scrCopy );

		WinSetDrawWindow( savewin );

		status_Show( *winptr, util_GetResourceString( ID_STRING_print_start ) );
	}

}

/*======================================================================
:::
::: status_Terminate()
:::
======================================================================*/
void status_Terminate( WindowType *window )
{
	RectangleType rect;

	if (window) {
		rect.topLeft.x = INFO_x;	
		rect.topLeft.y = INFO_y;	
		rect.extent.x  = INFO_w;	
		rect.extent.y  = INFO_h;	

		WinCopyRectangle( window, WinGetDisplayWindow(), &rect, 0,0, scrCopy );
		WinDeleteWindow( window, false );
	}
}

/*======================================================================
:::
::: ui_BusyWait()
:::
======================================================================*/
int ui_BusyWait( char *hold_flag, long mintime, long timeout )
{
	UInt error;
	long time1,time2;
	EventType event;

	time1 = TimGetTicks();

	do {

		EvtGetEvent(&event, 100);

		if (event.eType == appStopEvent) return -1;

		SysHandleEvent(&event);

		time2 = TimGetTicks();
next:
	} while (time2 >= time1 && time2 < time1+timeout && (time2<time1+mintime || !hold_flag || *hold_flag));

	if (time2 >= time1+timeout) return -2;
	return 0;
}



/*======================================================================
:::
::: PRINTING CODE
:::
======================================================================*/

	static Err	transfer_Initialize	( void **thandle, SavePrefsType *prefs );
	static void transfer_Terminate	( void *thandle );
	static Err	transfer_SendByte		( void *thandle, char data );
	static Err	transfer_Send			( void *thandle, long len, char *data );

/*======================================================================
:::
::: PrintTest()
:::
::: Test the printer by dumping some text
:::
======================================================================*/

long PrintTest( long mode )
{
	void *thandle;
	char *tptr;
	SavePrefsType prefs;
   char text[100];

	prefs_Load( &prefs );

	StrCopy( text, util_GetResourceString( ID_STRING_test0 ));
	if (prefs.linebreak !=2) StrCat( text, "\015" );
	if (prefs.linebreak !=1) StrCat( text, "\012" );
	if (prefs.linebreak !=2) StrCat( text, "\015" );
	if (prefs.linebreak !=1) StrCat( text, "\012" );

	StrCat( text, util_GetResourceString( ID_STRING_test1 ));
	if (prefs.linebreak !=2) StrCat( text, "\015" );
	if (prefs.linebreak !=1) StrCat( text, "\012" );
	if (prefs.linebreak !=2) StrCat( text, "\015" );
	if (prefs.linebreak !=1) StrCat( text, "\012" );

	StrCat( text, util_GetResourceString( ID_STRING_test2 ));
	if (prefs.linebreak !=2) StrCat( text, "\015" );
	if (prefs.linebreak !=1) StrCat( text, "\012" );

	StrCat( text, util_GetResourceString( ID_STRING_test3 ));
	if (prefs.linebreak !=2) StrCat( text, "\015" );
	if (prefs.linebreak !=1) StrCat( text, "\012" );
	StrCat( text, "\014" );


	if (!transfer_Initialize( &thandle, &prefs)) {

	 	transfer_Send( thandle, StrLen(text), text );

		SysTaskDelay(10);
		transfer_Terminate( thandle );
	}
}


/*======================================================================
:::
::: PrintEjectPaper()
:::
======================================================================*/

long PrintEjectPaper( void )
{
	SavePrefsType prefs;
	void *thandle;

	prefs_Load( &prefs );

	if (!transfer_Initialize( &thandle, &prefs)) {

		transfer_SendByte( thandle, 12 );
		transfer_Terminate( thandle );
	}
	return 0;
}

/*======================================================================
:::
::: PrintTextData()
:::
::: Print some text in native printer text format
::: This code should be almost the same for any
::: printer that supports native text
:::
======================================================================*/

long PrintTextData( int cols, int lines, int blank_cols, int blank_lines, void get_line( char **data, int line, int *offset, int *char_count, void *app_data ), void *app_data, long print_flags )
{
	int i,j;
	int count,offset,min_cols;
	void *thandle;
	unsigned char *tptr;
	SavePrefsType prefs;
	int aborted = true;
	char *text_remap;

	prefs_Load( &prefs );

	text_remap = prefs.charset ? util_GetResourceString( ID_STRING_text_remap0+prefs.charset-1 ):NULL;

	if (!transfer_Initialize( &thandle, &prefs)) {

		aborted = false;

		//
		// subtract off minimum margins
		//

		min_cols    = X_CPI * prefs.margin_left/1000;
		blank_lines = max( blank_lines-Y_CPI * prefs.margin_top/1000, 0 );
		cols = min( cols, X_CPI * MAX_PAPER_width/1000 );

		//
		// skip blank lines at the top of the page
		//

		for (i=0; i<blank_lines; i++) {
			if (prefs.linebreak !=2) if (transfer_SendByte( thandle, 13 )) goto fail;
			if (prefs.linebreak !=1) if (transfer_SendByte( thandle, 10 )) goto fail;
		}

		//
		// draw all the lines of data
		//

		for (i=0; i<lines; i++) {

			if (KeyCurrentState()) {
				aborted = true;
				break;		// abort on key press
			}

			((*get_line)( (char **)&tptr, 
							  blank_lines + Y_CPI * prefs.margin_top/1000 + i, 
							  &offset, &count, app_data ));

			if (tptr) {

				if (count) {

					// clip offset and count to small page width

					offset = min( offset, X_CPI * MAX_PAPER_width/1000 - min_cols );
					count  = min( count,  X_CPI * MAX_PAPER_width/1000 - offset   );

					// Send spaces for left margin

					if (offset >= min_cols) {
						for (j=0; j<offset-min_cols; j++) if (transfer_SendByte( thandle, ' ' )) goto fail;
						offset = 0;
					} else {
						offset = min_cols - offset;
					}

					// send body text

					if (count > offset) {
						for (j=0; j<count-offset; j++) {
							if (!text_remap || tptr[offset+j] < 127) {
								if (transfer_SendByte( thandle, tptr[offset+j] )) goto fail;
							} else {
								if (transfer_SendByte( thandle, text_remap[tptr[offset+j]-127] )) goto fail;
							}
						}
					}
				}

				if (prefs.linebreak !=2) if (transfer_SendByte( thandle, 13 )) goto fail;
				if (prefs.linebreak !=1) if (transfer_SendByte( thandle, 10 )) goto fail;
			}
		}

		// close everything off

		if (!(print_flags & DRIVER_OUTFLAG_noeject)) {
			if (transfer_SendByte( thandle, 12 )) goto fail;
		}
fail:
		SysTaskDelay(10);
		transfer_Terminate( thandle );
	}
	return aborted;
}

/*======================================================================
:::
::: PrintBitmapData()
:::
::: Print some bitmap data, given a callback function to get lines
::: of pixels
:::
::: This routine is highly printer specific.
::: This code is designed for the epson, which accepts data 8-lines at a 
::: time to work with original 9-pin epson dot matrix printers.
::: Most modern printers probably use data that is packed horizontally,
::: one row at a time.
:::
======================================================================*/

long PrintBitmapData( int width, int height, int blank_cols, int blank_lines, void (*get_line)( unsigned char **data, int line, int pixel_offset, int pixel_count, void *app_data, unsigned char bgcolor ), void *app_data, long flags )
{
	int i,j,k,c1,c2,mask;

	void *phandle;
	unsigned char *tptr;
	SavePrefsType prefs;
	unsigned char *tbuf;
	long linewidth = MAX_PIXEL_BUFFER;
	long workwidth;
	int aborted = true;

	prefs_Load( &prefs );

	width  += blank_cols;
	height += blank_lines;

	if (!transfer_Initialize( &phandle, &prefs )) {

		if (tbuf = MemPtrNew( 8 * linewidth )) {

			aborted = false;

			// wait for page feed (in case no handshaking)
										  
			SysTaskDelay( 50 );

			// subtract off minimum margins

			blank_cols  = 0;
			blank_lines = max( blank_lines-(long)Y_DPI * prefs.margin_top/1000, 0 );
			width = min( width, (long)X_DPI * MAX_PAPER_width/1000 );

			// Set line spacing 8/72"

			if (transfer_SendByte( phandle, 27  )) goto fail;	// reset
			if (transfer_SendByte( phandle, '@' )) goto fail;

			if (transfer_SendByte( phandle, 27  )) goto fail;	// v spacing
			if (transfer_SendByte( phandle, 'A' )) goto fail;
			if (transfer_SendByte( phandle, 8   )) goto fail;

			blank_lines &= ~7L;

			for (i=0; i<blank_lines/8; i++) {
				if (prefs.linebreak !=2) if (transfer_SendByte( phandle, 13 )) goto fail;
				if (prefs.linebreak !=1) if (transfer_SendByte( phandle, 10 )) goto fail;
			}

			width  -= blank_cols;
			height -= blank_lines;

			// Send data bytes

			for (i=0; i<height; i+=8) {

				if (KeyCurrentState()) {
					aborted=true;
					break;		// abort on key press
				}

				// Calc width (GCC trashes these calculations if imbedded in the function call

				c1 = width & 255;
				c2 = width>>8;

				// Send graphics command

				if (transfer_SendByte( phandle, 27 )) goto fail;
				if (transfer_SendByte( phandle, 'K')) goto fail;
				if (transfer_SendByte( phandle, c1 )) goto fail;
				if (transfer_SendByte( phandle, c2 )) goto fail;

				// Get 8 lines of pixels.  On High res printers, make sure we never 
				// ask for more than MAX_PIXEL_BUFFER bytes of data.
				for (k=0; k<width; k+=linewidth) {

					//
					// workwidth is count of pixels to fetch
					//

					workwidth = MIN( (width-k), linewidth );

					for (j=0; j<8; j++) {

						((*get_line)( &tptr, 
	 						  blank_lines + (long)Y_DPI * prefs.margin_top /1000 + i+j,
 							  blank_cols  + (long)X_DPI * prefs.margin_left/1000 + k,
							  workwidth, app_data, 0 ));

						if (tptr) {
							MemMove( &tbuf[linewidth * j+(k>>3)], tptr, (workwidth+7)/8 );
						}

						#if 0
							tbuf[linewidth*j  ] = j;
							tbuf[linewidth*j+1] = i;
						#endif
					}

					mask = 128;
					tptr = tbuf;

					for (j=0; j<workwidth; j++) {

						c1 = 0;
						if (tptr[ 0                                ] & mask) c1 |= 128;
						if (tptr[ linewidth                        ] & mask) c1 |= 64;
						if (tptr[ linewidth<<1                     ] & mask) c1 |= 32;
						if (tptr[ linewidth+linewidth+linewidth    ] & mask) c1 |= 16;
						if (tptr[ linewidth<<2                     ] & mask) c1 |= 8;
						if (tptr[(linewidth<<2)+linewidth          ] & mask) c1 |= 4;
						if (tptr[(linewidth+linewidth+linewidth)<<1] & mask) c1 |= 2;
						if (tptr[(linewidth<<3)-linewidth          ] & mask) c1 |= 1;

						if (transfer_SendByte( phandle, c1)) goto fail;

						mask >>= 1;
						if (!mask) {
							mask = 128;
							tptr++;
						}
					}
				}

				if (prefs.linebreak !=2) if (transfer_SendByte( phandle, 13 )) goto fail;
				if (prefs.linebreak !=1) if (transfer_SendByte( phandle, 10 )) goto fail;
			}
			MemPtrFree( tbuf );
		}


		// close everything off

		if (transfer_SendByte( phandle, 27  )) goto fail;
		if (transfer_SendByte( phandle, '2' )) goto fail;

		if (!(flags & DRIVER_OUTFLAG_noeject)) {
			SysTaskDelay( 10 );
			if (transfer_SendByte( phandle, 12 )) goto fail;
		}
fail:
		SysTaskDelay(10);
		transfer_Terminate( phandle );
	}
	return aborted;
}

/*======================================================================
:::
::: COMMUNICATION CODE
:::
======================================================================*/

//
// SERIAL
//

	#include <System/SerialMgr.h>
	#include <System/dlserver.h>

	#define ID_CREATOR_psys	'psys'
	#define ID_FTR_osversion	1

	typedef struct {
		int	mode;
		WindowType *infowin;

		UInt	serial_ref;
		UInt	serial_open;
		UInt	serial_timeout;
		UInt	serial_send_okay;

		SerSettingsType serial_settings;
		Boolean	send_slow;

		int	beep_status;
		long	offset;

	} SerialGlobalsType;

	static Err  	serial_Initialize		( SerialGlobalsType *ss, long baud_rate, int rts_cts, int xon_xoff, int parity, Boolean send_slow );
	static void 	serial_Terminate 		( SerialGlobalsType *ss );
	static Err		serial_Send  			( SerialGlobalsType *ss, long len, char *data );
	static int 		serial_Receive			( SerialGlobalsType *ss, char *buffer, long *offset, long bufsize );
	static int 		serial_ReceiveWait	( SerialGlobalsType *ss, long bytes, int timeout );
	static Err 		serial_SendByte		( SerialGlobalsType *ss, char data );

//
// IRDA
//

	#include <irlib.h>

	#define RESULT_SIZE	128

	typedef struct {
		int	mode;
		WindowType *infowin;

		UInt ir_ref;

		IrConnect con;
		IrPacket packet;
		IrIasQuery query;
		IrIasAttribute attribs;
		IrIasObject ias_object;

		int current_state;
		long counter;
		char	result[RESULT_SIZE];
		char attr_value[10];

		char 			 init_wait;			// 1) Initialize

		char 			 discover_wait;	// 2) Discover
		IrDeviceAddr discover_addr;
		char 			 discover_success;

		char 			 irlap_wait;		// 3) Connect/Disconnect IrLap
		char 			 irlap_success;

		char 			 irlmp_wait;		// 4) Query Lsap, chains to connect to irlmp
		char 			 irlmp_success;

		char 			 datareq_wait;		// 5) DataReq
		long 			 datareq_size;

		int	beep_status;
		long	offset;

	} IrDaState;

	int	irda_Discover			( IrDaState *is ); 
	int	irda_Shutdown			( IrDaState *is ); 
	int	irda_Startup			( IrDaState *is, long speed );
	void	irda_Callback			( IrConnect *connect, IrCallBackParms *parms );
	int	irda_ConnectIrLap		( IrDaState *is );
	int	irda_DisconnectIrLap	( IrDaState *is );
	int	irda_ConnectIrLmp		( IrDaState *is );
	int	irda_DataReq			( IrDaState *is, char *data, int len );


/*======================================================================
:::
::: transfer_Initialize()
:::
======================================================================*/

static Err transfer_Initialize( void **thandle, SavePrefsType *prefs )
{
	//==================================================================
	//
	// Infared Code
	//
	//==================================================================

	if (prefs->infared) {
		IrDaState *is;
		WindowType *win;

		if (is = MemPtrNew( sizeof(*is) )) {

			//
			// Initialize IrDA system and establish a full connection
			//

			status_Initialize( &win, prefs );

			if (!irda_Startup(is, prefs->baud_rate)) {

				//
				// Discover Neighbors
				//
	
				if (!ui_BusyWait( &is->init_wait, 200,500 ) && !irda_Discover(is)) {
					status_Show( win, util_GetResourceString( ID_STRING_ir_discover ));
	
					if (!ui_BusyWait( &is->discover_wait, 200,500 ) && is->discover_success) {
	
						//
						// Connect IrLap services
						//
	
					 	if ( !irda_ConnectIrLap(is)) {
							status_Show( win, util_GetResourceString( ID_STRING_ir_irlap ));
	
							if (!ui_BusyWait( &is->irlap_wait, 300,500 ) && is->irlap_success) {

								//
								// Piggy back IrLmp services
								//
	
								if ( !irda_ConnectIrLmp(is)) {
									status_Show( win, util_GetResourceString( ID_STRING_ir_irlmp ));
		
									if (!ui_BusyWait( &is->irlmp_wait, 300,1000 )) {

										is->datareq_size = IrMaxTxSize(is->ir_ref, &is->con);

										status_Show( win, util_GetResourceString( ID_STRING_ir_datareq ));

										//
										// Ready for data.
										//

										is->mode = 1;
										is->infowin = win;
										is->beep_status = prefs->beep_status;

										*thandle = (void *)is;
										return 0;
									} else {
										status_Show( win, util_GetResourceString( ID_STRING_ir_timeout ));
									}
								} else {
									status_Show( win, util_GetResourceString( ID_STRING_ir_noconn ));
								}

								if (!irda_DisconnectIrLap(is)) {
									status_Show( win, util_GetResourceString( ID_STRING_ir_disconnect ));
									ui_BusyWait( &is->irlap_wait, 100,500 );
								}
							}
						} else {
							status_Show( win, util_GetResourceString( ID_STRING_ir_busy ));
						}
					} else {
						status_Show( win, util_GetResourceString( ID_STRING_ir_noreply ));
					}
				} else {
					status_Show( win, util_GetResourceString( ID_STRING_ir_noport ));
				}

			 	irda_Shutdown(is);
			}

			FrmAlert( ID_ALERT_error_timed_out );

			status_Terminate( win );
			MemPtrFree( is );
		}
		return -1;
	} 

	//==================================================================
	//
	// Serial Code
	//
	//==================================================================

	else {
		SerialGlobalsType *ss;

		if (ss = MemPtrNew( sizeof(*ss) )) {
			MemSet( ss, sizeof(*ss), 0 );

			if (!serial_Initialize( ss, prefs->baud_rate, prefs->rts_cts, prefs->xon_xoff, prefs->parity, prefs->sendslow )) {

				status_Initialize( &ss->infowin, prefs );

				ss->beep_status = prefs->beep_status;
				*thandle = (void *)ss;
				return 0;
			}
			MemPtrFree( ss );
		}
		return -1;
	}
}

/*======================================================================
:::
::: transfer_Terminate()
:::
======================================================================*/

static void transfer_Terminate( void *thandle )
{
	//
	// Infared Code
	//

	if ( *(int *)thandle ) {
 
		IrDaState *is = (IrDaState *)thandle;

		if (!irda_DisconnectIrLap(is)) {
			status_Show( is->infowin, util_GetResourceString( ID_STRING_ir_disconnect ));
			ui_BusyWait( &is->irlap_wait, 100,500 );
		}

	 	irda_Shutdown(is);

		status_Terminate( is->infowin );
		MemPtrFree( is );
	}

	//
	// Serial Code
	//

	else {
		SerialGlobalsType *ss = (SerialGlobalsType *)thandle;

		serial_Terminate( ss );

		status_Terminate( ss->infowin );
		MemPtrFree( ss );
	}
}

/*======================================================================
:::
::: transfer_SendByte()
:::
======================================================================*/

static Err transfer_SendByte( void *thandle, char data )
{
	return transfer_Send( thandle,1,&data );
}

/*======================================================================
:::
::: transfer_Send()
:::
======================================================================*/

static Err transfer_Send( void *thandle, long len, char *data )
{
	//
	// Infared Code
	//

	if ( *(int *)thandle ) {
 
		IrDaState *is = (IrDaState *)thandle;

		while ( len > is->datareq_size ) {

			irda_DataReq( is, data, is->datareq_size );
			if (ui_BusyWait( &is->datareq_wait, 0,1000 )) return -1;

			data += is->datareq_size;
			len  -= is->datareq_size;
		}

		irda_DataReq( is, data, len );
		return ui_BusyWait( &is->datareq_wait, 0,1000 );
	}

	//
	// Serial Code
	//

	else {
		return serial_Send( thandle, len, data );
	}
}


/*======================================================================
:::
::: serial_Initialize()
:::
::: Open the serial library and initialize the port
:::
======================================================================*/

	#define SERIAL_TIMEOUT	150	// initial timeout
	#define SERIAL_TIMEOUT2	50		// running timeout

static Err serial_Initialize( SerialGlobalsType *ss, long baud_rate, int rts_cts, int xon_xoff, int parity, Boolean sendslow )
{
	Err error;


	ss->send_slow = sendslow;
	ss->serial_open = 0;
	ss->serial_send_okay = true;

	//
	// Set Serial Port Settings
	//

	ss->serial_settings.baudRate   = baud_rate;
	ss->serial_settings.ctsTimeout = ss->serial_timeout = SERIAL_TIMEOUT;
	ss->serial_settings.flags 	   = 
										  serSettingsFlagStopBits1 	| 
										  serSettingsFlagBitsPerChar8 | 
									  (  parity     ?	  serSettingsFlagParityOnM   : 0 ) | 
									  ( (parity==2) ?	  serSettingsFlagParityEvenM : 0 ) | 
									  (  xon_xoff   ?	  serSettingsFlagXonXoffM    : 0 ) | 
									  (  rts_cts    ? ( serSettingsFlagCTSAutoM )  : 0 ) |
											0;

	if (!ss->serial_settings.baudRate) return 0;

	//
	// Try to open the serial port...
	//

	error = SysLibFind("Serial Library", &ss->serial_ref);

	if (error) {
		FrmAlert( ID_ALERT_error_find );
	} 

	else {
		error = SerOpen( ss->serial_ref, 0, baud_rate );
	
		if (error) {

			if( error == serErrAlreadyOpen)    
				FrmAlert( ID_ALERT_error_in_use );
			if( error == memErrNotEnoughSpace) 
				FrmAlert( ID_ALERT_out_of_memory );

			return error;
		}
	}
	ss->serial_open = 1;
					
	error = SerSetSettings(ss->serial_ref, &ss->serial_settings);

	if (error) {
		SerClearErr( ss->serial_ref );	
		FrmAlert( ID_ALERT_error_settings );
	}

	return 0;
}

/*======================================================================
:::
::: serial_Terminate()
:::
::: Close the serial library
:::
======================================================================*/


static void serial_Terminate( SerialGlobalsType *ss )

{
	if (!ss->serial_settings.baudRate) return;

	if (ss->serial_open) {
		SerClose(ss->serial_ref);
		ss->serial_open = 0;
		if (ss->beep_status)	SndPlaySystemSound( sndInfo );
	}        
}

/*======================================================================
:::
::: serial_SendByte()
:::
======================================================================*/
static Err serial_SendByte( SerialGlobalsType *ss, char data )
{
	int error;
	error = serial_Send( ss,1,&data );
	SysTaskDelay(1);
	return error;
}

/*======================================================================
:::
::: serial_Send()
:::
::: Send a block out the serial port
:::
======================================================================*/

static Err serial_Send( SerialGlobalsType *ss, long len, char *data )
{
	Err error = 0;

	//
	// Output Infared bytes count... update every 4k
	//

	ss->offset += len;
	if ( (ss->offset & 0xfffff800 ) != ((ss->offset-len) & 0xfffff800 )) {
		int slen;
		char str[50];

		if (ss->infowin) {
			StrCopy( str, util_GetResourceString( ID_STRING_printk ));
			slen = StrLen( str );
			StrIToA( str+slen, ss->offset );
			status_Show( ss->infowin, str );
		}
		if (ss->beep_status) {
			SndPlaySystemSound( sndClick );
		}
	}

	//
	// Send Stuff
	//

	if (!ss->serial_settings.baudRate) return 0;

	if (ss->send_slow) {
		error = 0;
		while (len-- && !error) {
			if (error = SerSend10( ss->serial_ref, data++, 1 )) {
				if ( ss->serial_send_okay ) {
					FrmAlert( ID_ALERT_error_timed_out );
				}

				ss->serial_send_okay = false;
				SerClearErr( ss->serial_ref );	
				return -1;
			}

			SysTaskDelay(1);
			ss->serial_send_okay = true;
			SerSendWait( ss->serial_ref, -1);	
		}
	} 

	else {
		error = SerSend10( ss->serial_ref, data, len );

		if (error) {
			// flag first error
	
			if ( ss->serial_send_okay ) {
				FrmAlert( ID_ALERT_error_timed_out );
			}
	
			ss->serial_send_okay = false;
			SerClearErr( ss->serial_ref );	
			return -1;
		}
	
		else {
	
			ss->serial_send_okay = true;
	
			SerSendWait( ss->serial_ref, -1);	
	
			#if 0
			if ( ss->serial_timeout != SERIAL_TIMEOUT2 ) {
				SerSetSettings(ss->serial_ref, &ss->serial_settings);
				ss->serial_settings.ctsTimeout = ss->serial_timeout = SERIAL_TIMEOUT2;
				SerSetSettings(ss->serial_ref, &ss->serial_settings);
			}
			#endif

		}
	}
	
	return error;
}

/*======================================================================
:::
::: serial_Receive()
:::
::: Get a block in the serial port
:::
======================================================================*/

static int serial_Receive( SerialGlobalsType *ss, char *buffer, long *offset, long bufsize )
{
	Err error;
	ULong num_bytes = 0;			

	if (!ss->serial_settings.baudRate) return 0;

	if (!ss->serial_open) return -1;

	//
	// Check for Data
	//

	error = SerReceiveCheck(ss->serial_ref, &num_bytes);

	if (error) {

		SerClearErr(ss->serial_ref);
		FrmAlert( ID_ALERT_error_checking_port );

		return error;
	}

	//
	// Get Data
	//

	if (num_bytes <= 0) return -2;
	
	if (num_bytes+*offset > bufsize) num_bytes = bufsize-*offset;

	if (num_bytes <= 0) return 0;

	error = SerReceive10(ss->serial_ref, buffer+*offset, num_bytes, -1);	

	if (error) {

		switch( error ) {

			case serErrTimeOut:
				FrmAlert( ID_ALERT_error_timed_out );
				break;

			case serErrLineErr:
				FrmAlert( ID_ALERT_error_line_error );
				break;

			default:
				FrmAlert( ID_ALERT_error_read_error );
				break;
		}

		SerClearErr(ss->serial_ref);

		return error;
	}

	*offset += num_bytes;
							
	return 0;
}

/*======================================================================
:::
::: serial_ReceiveWait()
:::
======================================================================*/

static int serial_ReceiveWait( SerialGlobalsType *ss, long bytes, int timeout )
{
	Err error;

	if (!ss->serial_settings.baudRate) return 0;

	if (error = SerReceiveWait(ss->serial_ref, bytes, timeout )) {
		SerClearErr( ss->serial_ref );	
	}
	return error;
}

//======================================================================
//
// IRDA Code
//
//======================================================================

	void irda_Callback( IrConnect *connect, IrCallBackParms *parms );
	void irda_IasCallback( IrStatus status );

	#define S_IAS_NAME "TealPrint"
	#define N_IAS_NAME 9

	#define S_DEV_NAME "\002\0PalmE"
	#define N_DEV_NAME 7

	#define S_ATTR_NAME "IrDA:TinyTP:LsapSel"
	#define N_ATTR_NAME 19

/*======================================================================
:::
::: irda_Startup()
:::
======================================================================*/

int irda_Startup( IrDaState *is, long baud_rate ) 
{
	IrStatus ret;

	// save the pointer to our state structure for the callback function

	FtrSet( CREATOR_ID, ID_FTR_irda_state, (long)is );

	MemSet( is, sizeof(*is), 0 );

	is->ir_ref = 0;

	// find IR library

	if (SysLibFind(irLibName, &is->ir_ref)) return -1;

	// open IR library

	if ( !IrOpen(is->ir_ref, (baud_rate==9600) ? irOpenOptSpeed9600 : (baud_rate==115200) ? irOpenOptSpeed115200 : 0) ) {

		// bind us

	  	if (IrBind(is->ir_ref, &is->con, irda_Callback)==IR_STATUS_SUCCESS) {

			if (IrSetDeviceInfo(is->ir_ref, S_DEV_NAME, N_DEV_NAME)==IR_STATUS_SUCCESS) {

				is->ias_object.name     = S_IAS_NAME;
				is->ias_object.len      = N_IAS_NAME;
				is->ias_object.nAttribs = 1;
				is->ias_object.attribs  = &is->attribs;

				is->attribs.name   = S_ATTR_NAME;     /* Pointer to name of attribute */
				is->attribs.len    = N_ATTR_NAME;     /* Length of attribute name */
				is->attribs.value  = is->attr_value;  /* Hardcode value (see below) */
				is->attribs.valLen = 5; 				  /* Length of the value. */

				is->attr_value[0] = IAS_ATTRIB_INTEGER;
				is->attr_value[1] = 0;
				is->attr_value[2] = 0;
				is->attr_value[3] = 0;
				is->attr_value[4] = is->con.lLsap;

				// set up our service in IAS db

				is->init_wait = true;
				if ((ret=IrIAS_Add(is->ir_ref, &is->ias_object))==IR_STATUS_SUCCESS) {
					return 0;
				}
			}
		}

		IrClose(is->ir_ref);
	}

	is->ir_ref = 0;
	return -1;
}

/*======================================================================
:::
::: irda_Shutdown()
:::
======================================================================*/

int irda_Shutdown( IrDaState *is ) 
{
	if (is->ir_ref) {
		IrClose(is->ir_ref);
		is->ir_ref = 0;
		if (is->beep_status)	SndPlaySystemSound( sndInfo );
	}
}

/*======================================================================
:::
::: irda_Discover()
:::
======================================================================*/

int irda_Discover( IrDaState *is ) 
{
	if (IrIsMediaBusy(is->ir_ref)) return -1;

	// search for someone

	is->discover_success = false;
	if ((IrDiscoverReq(is->ir_ref, &is->con))==IR_STATUS_PENDING) {
		is->discover_wait = true;
		return 0;
	}

	return -2;
}

/*======================================================================
:::
::: irda_ConnectIrLap()
:::
======================================================================*/

int irda_ConnectIrLap( IrDaState *is ) 
{
	if (!is->discover_success) return -1;

	// search for someone

	is->irlap_success = false;

	if ( IrConnectIrLap(is->ir_ref, is->discover_addr )==IR_STATUS_PENDING) {
		is->irlap_wait = true;
		return 0;
	}

	return -2;
}

/*======================================================================
:::
::: irda_DisconnectIrLap()
:::
======================================================================*/

int irda_DisconnectIrLap( IrDaState *is ) 
{
	if (!IrIsIrLapConnected(is->ir_ref)) return -1;

	// search for someone

	if ( IrDisconnectIrLap(is->ir_ref)==IR_STATUS_PENDING) {
		is->irlap_wait = true;
		return 0;
	}

	return -2;
}

/*======================================================================
:::
::: irda_ConnectIrLmp()
:::
======================================================================*/

int irda_ConnectIrLmp( IrDaState *is ) 
{
	IrSetConTypeLMP(&is->con);

	// get rLsap

	is->query.queryBuf = "\005IrLPT\022IrDA:IrLMP:LsapSel";

	is->query.queryLen = 25;
	is->query.resultBufSize = RESULT_SIZE;
	is->query.result = is->result;
	is->query.callBack = irda_IasCallback;
	is->irlmp_success = false;
	
	if ( IrIAS_Query(is->ir_ref, &is->query)==IR_STATUS_PENDING ) {
		is->irlmp_wait = 1;
		return 0;
	}

	return -2;
}									

/*======================================================================
:::
::: irda_DataReq()
:::
======================================================================*/

int irda_DataReq( IrDaState *is, char *data, int len ) 
{
	is->packet.buff = data;
	is->packet.len = len;
			
	if( IrDataReq(is->ir_ref, &is->con, &is->packet)==IR_STATUS_PENDING ) {
		is->datareq_wait = 1;

		//
		// Output Infared bytes count... update every 4k
		//

		is->offset += len;
		if ( (is->offset & 0xfffff800 ) != ((is->offset-len) & 0xfffff800 )) {
			int slen;
			char str[50];

			if (is->infowin) {
				StrCopy( str, util_GetResourceString( ID_STRING_printk ));
				slen = StrLen( str );
				StrIToA( str+slen, is->offset );
				status_Show( is->infowin, str );
			}
			if (is->beep_status) {
				SndPlaySystemSound( sndClick );
			}
  		}

		return 0;
	}

	return -2;
}									

/*======================================================================
:::
::: irda_Callback()
:::
======================================================================*/

void irda_Callback( IrConnect *connect, IrCallBackParms *parms ) 
{
	IrStatus ret;
	int i;
	IrDaState *is;

	// CALLBACK_PROLOGUE;

	FtrGet( CREATOR_ID, ID_FTR_irda_state, (long *)&is );

	switch(parms->event) {

		//
		// Initialized 
		//

		case LEVENT_STATUS_IND:
			if (is->init_wait) is->init_wait = false;
			break;

		//
		// Discovery
		//

		case LEVENT_DISCOVERY_CNF:

			if (is->discover_wait) {

				if(parms->deviceList->nItems > 0) {

					is->discover_addr.u32= parms->deviceList->dev[0].hDevice.u32;
					is->discover_success = true;
				} else {
					is->discover_success = false;
				}

				is->discover_wait = false;
			}
			break;

		//
		// IrLap Connect/Disconnect
		//

		case LEVENT_LAP_DISCON_IND:
			if (is->irlap_wait) {
				is->irlap_success = false;
				is->irlap_wait = false;
			}
			if (is->irlmp_wait) {
				is->irlmp_wait = false;
			}
			break;

		case LEVENT_LAP_CON_CNF:
			if (is->irlap_wait) {
				is->irlap_success = true;
				is->irlap_wait = false;
			}
			break;

		//
		// IrLmp connection
		//
		
		case LEVENT_LM_CON_IND:

			// accept connection

			is->packet.buff = "Test Data";
			is->packet.len = 0;

			if ((ret = IrConnectRsp( is->ir_ref, &is->con, &is->packet, 0x7f))==IR_STATUS_PENDING) {
				is->irlmp_wait++;
			} else {
				is->irlmp_wait = 0;
			}
			break;

		case LEVENT_LM_CON_CNF:
			is->irlmp_success = true;
			is->irlmp_wait = 0;
			break;

		case LEVENT_DATA_IND:
			break;

		case LEVENT_PACKET_HANDLED:
			if (is->datareq_wait) is->datareq_wait = false;
			break;
	}
	// CALLBACK_EPILOGUE;
}

/*======================================================================
:::
::: irda_IasCallback()
:::
::: Callback used to return information queried from remote device's
::: IAS object database
:::
======================================================================*/

void irda_IasCallback( IrStatus status )
{
	IrStatus ret;
	Byte b;
	int i;
	IrDaState *is;

	// CALLBACK_PROLOGUE;

	FtrGet( CREATOR_ID, ID_FTR_irda_state, (long *)&is );

	if((is->query.retCode)==IAS_RET_SUCCESS) {

		i = IrIAS_GetType(&is->query);

		switch(i) {

			//
			// Got integer returned (like remote LSAP)
			//

			case IAS_ATTRIB_INTEGER:

				//
				// Got Lsap query.  Now send out LMP connect
				//
				if (is->irlmp_wait==1) {
					is->irlmp_wait++;

					is->con.rLsap = IrIAS_GetIntLsap(&is->query);

					// build my first IrPacket and buffer

					is->packet.buff = "Test Data";
					is->packet.len = 0;
		
					// open session using the lsap we just got

					if ( (ret = IrConnectReq(is->ir_ref, &is->con, &is->packet, 0x7f))!=IR_STATUS_PENDING) {
						is->irlmp_wait = 0;
					}
				}
				break;

			//
			// Got string returned
			//

			case IAS_ATTRIB_USER_STRING:
				break;
			
			case IAS_ATTRIB_OCTET_STRING:
				break;
			
			case IAS_ATTRIB_MISSING:
			default:
				break;
			}
	}

	// CALLBACK_EPILOGUE;
}

