/*======================================================================
:::
::: util_DoDialogCallback()
:::
::: Just like FrmDoDialog(), except pass a callback routine that
::: returns 1 for handled
::: sets retval to close dialog
:::
======================================================================*/

int util_DoDialogCallback( FormPtr form, int (*callback)( EventType *evt, void *data, int *retval), void *data )
{
	FormPtr oldform;
	EventType evt;	
	int retval = 0;
	UInt err;

	oldform = FrmGetActiveForm();

	FrmSetActiveForm( form );
	FrmDrawForm( form );

	do {
		EvtGetEvent( &evt, evtWaitForever );

		if (callback) if ((*callback)(&evt,data,&retval)) continue;
 
		if (SysHandleEvent(&evt)) continue;
		if (MenuHandleEvent((void *)0, &evt, &err)) continue;

		switch (evt.eType) {

			case ctlSelectEvent:
				retval = evt.data.ctlSelect.controlID;
				break;
		
   		case nilEvent:
			default:
				FrmHandleEvent(form, &evt);
				break;

			case appStopEvent:
				EvtAddEventToQueue( &evt );
				retval = -1;
				break;

			case frmCloseEvent:
				retval = -1;
				break;

		}
	} while ( !retval );

	FrmEraseForm( form );
	FrmSetActiveForm( oldform );

//	EvtFlushPenQueue();
	return retval;
}

/*------------------------------------------------
:::
::: util_GetControlValue()
:::
------------------------------------------------*/
int util_GetControlValue( FormPtr form, int index )
{
	return FrmGetControlValue( form, FrmGetObjectIndex( form, index ) );
}

/*------------------------------------------------
:::
::: util_GetObjectPtr()
:::
------------------------------------------------*/
void *util_GetObjectPtr( FormPtr form, int index )
{
	return FrmGetObjectPtr( form, FrmGetObjectIndex( form, index ) );
}

/*------------------------------------------------
:::
::: util_SetControlValue()
:::
------------------------------------------------*/
void util_SetControlValue( FormPtr form, int index, int value )
{
	FrmSetControlValue( form, FrmGetObjectIndex( form, index ), value );
}

/*------------------------------------------------
:::
::: util_SetTriggerLabel()
:::
------------------------------------------------*/
void util_SetTriggerLabel( FormPtr form, int id, int list, int index )
{
	CategorySetTriggerLabel( util_GetObjectPtr( form, id ), LstGetSelectionText( util_GetObjectPtr(form,list),index) );
	LstSetSelection( util_GetObjectPtr(form,list),index );
}

/*------------------------------------------------
:::
::: util_AllocStringHandle()
:::
------------------------------------------------*/
VoidHand util_AllocStringHandle( char *text )
{
	VoidHand handle;
	VoidPtr  ptr;

	if (handle = MemHandleNew(StrLen(text)+1)) {
		if (ptr = MemHandleLock(handle)) {

			StrCopy( ptr,text );
			MemHandleUnlock(handle);
		}
	}
	return handle;
}

/*------------------------------------------------
:::
::: util_AllocNumberHandle()
:::
------------------------------------------------*/
VoidHand util_AllocNumberHandle( int number, int decimals )
{
	VoidHand handle;
	char *   ptr;
	int len,i;

	if (handle = MemHandleNew(20)) {
		if (ptr = MemHandleLock(handle)) {

			StrIToA( ptr,number );

			if (decimals) {
				len = StrLen( ptr );

				// 123456
				// 123.456

				if (len>decimals) {
					MemMove( &ptr[len+1-decimals], &ptr[len-decimals], decimals+1 );
					ptr[len-decimals] = '.';
				} 

				// 12
				// 0.012

				else {
					MemMove( &ptr[decimals+2-len], &ptr[0], decimals+1 );
					ptr[0] = '0';
					ptr[1] = '.';
					MemSet( &ptr[2], decimals-len, '0' );
				}
			}

			MemHandleUnlock(handle);
		}
	}
	return handle;
}

/*------------------------------------------------
:::
::: util_SetFieldNumber()
:::
------------------------------------------------*/
void util_SetFieldNumber( FormPtr form, int id, int number, int decimals, int drawit )
{
	Handle handle;
	FieldPtr field;
	
	if (handle = FldGetTextHandle( field = util_GetObjectPtr( form, id ))) {
		FldSetTextHandle( field, 0 );
		MemHandleFree( (VoidHand)handle );
	}

	FldSetTextHandle( field, (void *)util_AllocNumberHandle(number, decimals) );
	if (drawit) FldDrawField( field );
}

/*------------------------------------------------
:::
::: util_GetFieldNumber()
:::
------------------------------------------------*/
int util_GetFieldNumber( FormPtr form, int id, int decimals )
{
	Handle handle;
	FieldPtr field;
	char *ptr;
	int retval = -1;
	int i,j,dec= -1;

	if (handle = FldGetTextHandle( field = util_GetObjectPtr( form, id ))) {
		if (ptr = MemHandleLock((VoidHand)handle)) {

			char text[14];

			// copy number, stripping off decimal

			for (i=0,j=0; i<9; j++) {
				if (ptr[j]=='.') {
					dec = decimals;
					decimals = 0;
				} else {
					if (!(text[i++] = ptr[j])) break;
	
					if (dec>=0) {
						if (!dec--) {
							text[i] = 0;
							break;
						}
					}
				}
			}
			while (dec-->0 || decimals-->0) {
				StrCat( text, "0" );
			}

			retval = StrAToI( text );
	
			MemHandleUnlock((VoidHand)handle);
		}
	}
	return retval;
}

/*------------------------------------------------
:::
::: util_GetResourceString()
:::
------------------------------------------------*/

char *util_GetResourceString( int id )
{
	VoidPtr handle;
	VoidPtr tptr;

	if (handle = DmGetResource( 'tSTR', id )) {
		if (tptr = MemHandleLock(handle)) {
			MemHandleUnlock(handle);
		}
		DmReleaseResource(handle);
	}
	return tptr;
}

/*======================================================================
:::
::: debug_Print()
:::
======================================================================*/

void debug_Print( long number, int x, int y)
{
	char debug_buffer[20];
   StrIToA( debug_buffer, number );
   WinDrawChars( debug_buffer, 1+StrLen(debug_buffer), x,y );
	SysTaskDelay(10);
}

void debug_Line( int col, int row )
{
	ULong **SSA = (ULong **)0xfffffA00;		// display hardware

	(*SSA)[row*5+col] = 0xaaaaaaaa; 
}

